import isArray from 'lodash/isArray'
import includes from 'lodash/includes'
import startsWith from 'lodash/startsWith'
import endsWith from 'lodash/endsWith'
import normalize from './normalize'

export default function(actual, operator, expected, el$, form$) {
  if (!operator) {
    return false
  }

  const strict = form$.$vueform.config.strictConditions

  actual = Array.isArray(actual) ? actual.map(e => normalize(e)) : normalize(actual)
  expected = Array.isArray(expected) ? expected.map(e => normalize(e)) : normalize(expected)

  const moment = form$.$vueform.services.moment
  
  switch (operator.toLowerCase()) {
    case '>':
      return isArray(actual)
        ? actual.every(a => a > expected)
        : actual > expected

    case '>=':
      return isArray(actual)
        ? strict
          ? actual.every(a => a >= expected && a !== null && a !== undefined && a !== '')
          : actual.every(a => a >= expected)
        : strict
          ? actual >= expected && actual !== null && actual !== undefined && actual !== ''
          : actual >= expected

    case '<':
      return isArray(actual)
        ? strict
          ? actual.every(a => a < expected && a !== null && a !== undefined && a !== '')
          : actual.every(a => a < expected)
        : strict
          ? actual < expected && actual !== null && actual !== undefined && actual !== ''
          : actual < expected

    case '<=':
      return isArray(actual)
        ? strict
          ? actual.every(a => a <= expected && a !== null && a !== undefined && a !== '')
          : actual.every(a => a <= expected)
        : strict
          ? actual <= expected && actual !== null && actual !== undefined && actual !== ''
          : actual <= expected

    case 'between':
      return actual > expected[0] && actual < expected[1]

    case 'empty':
      if (isArray(actual)) {
        return !actual.length
      } else if (actual && actual instanceof File) {
        return false
      } else if (actual && actual instanceof Blob) {
        return false
      } else if (actual && typeof actual === 'object') {
        let values = Object.values(actual)

        return !values.length || values.every((v) => {
          if (Array.isArray(v)) {
            return !v.length
          } else if (v && typeof v === 'object') {
            return Object.values(v).every((ov) => {
              if (Array.isArray(ov)) {
                return !ov.length
              } else {
                return ['', null, undefined].indexOf(ov) !== -1
              }
            })
          } else {
            return ['', null, undefined].indexOf(v) !== -1
          }
        })
      } else {
        return ['', null, undefined].indexOf(actual) !== -1
      }

    case 'not_empty':
      if (isArray(actual)) {
        return !!actual.length
      } else if (actual && actual instanceof File) {
        return true
      } else if (actual && actual instanceof Blob) {
        return true
      } else if (actual && typeof actual === 'object') {
        let values = Object.values(actual)

        return values.length && values.some((v) => {
          if (Array.isArray(v)) {
            return v.length
          } else if (v && typeof v === 'object') {
            return Object.values(v).some((ov) => {
              if (Array.isArray(ov)) {
                return ov.length
              } else {
                return ['', null, undefined].indexOf(ov) === -1
              }
            })
          } else {
            return ['', null, undefined].indexOf(v) === -1
          }
        })
      } else {
        return ['', null, undefined].indexOf(actual) === -1
      }

    case '==':
    case 'in':
      if (isArray(expected)) {
        if (isArray(actual)) {
          // ['checkboxes', [1,2,3]]
          return !expected.length
            ? !actual.length
            : actual.filter(a => includes(expected, a)).length > 0
        } else {
          // ['text', [1,2,3]]
          return expected.indexOf(actual) !== -1
        }
      } else {
        if (isArray(actual)) {
          // ['checkboxes', 1]
          return actual.indexOf(expected) !== -1
        } else {
          // ['text', 1]
          return actual == expected 
        }
      }

    case '!=':
    case 'not_in':
      if (isArray(expected)) {
        if (isArray(actual)) {
          // ['checkboxes', 'not_in', [1,2,3]]
          return !expected.length
            ? !!actual.length
            : actual.filter(e => includes(expected, e)).length == 0
        } else {
          // ['text', 'not_in', [1,2,3]]
          return expected.indexOf(actual) === -1
        }
      } else {
        if (isArray(actual)) {
          // ['checkboxes', '!=', 1]
          return actual.indexOf(expected) === -1
        } else {
          // ['text', '!=', 1]
          return actual != expected 
        }
      }

    case 'today':
      if (!isArray(actual)) {
        actual = [actual]
      }

      return actual.length && actual.every(a => moment(a, el$.valueDateFormat).isSame(moment(), 'day'))

    case 'before':
      if (!isArray(actual)) {
        actual = [actual]
      }

      return actual.length && actual.every((a) => {
        let date = moment(a, el$.valueDateFormat)

        return date.isValid() && date.isBefore(moment(expected === 'today' ? undefined : expected), 'day')
      })

    case 'after':
      if (!isArray(actual)) {
        actual = [actual]
      }

      return actual.length && actual.every((a) => {
        let date = moment(a, el$.valueDateFormat)

        return date.isValid() && date.isAfter(moment(expected === 'today' ? undefined : expected), 'day')
      })

    case '^':
      return startsWith(actual, expected)

    case '$':
      return endsWith(actual, expected)

    case '*':
      return includes(actual, expected)

    default:
      const customOperators = form$.$vueform.config.operators || {}

      if (customOperators[operator]) {
        return customOperators[operator](actual, expected, el$, form$)
      }

  }
}